// Copyright (C) 2009-2010 Ferdinand Majerech
// This file is part of MiniINI
// For conditions of distribution and use, see copyright notice in LICENSE.txt

#ifndef INISECTION_H_INCLUDED
#define INISECTION_H_INCLUDED

#include "typedefs.h"

///Section of ini file. 
/*
 * Stores tags of an ini file section.
 */
class INISection
{
    friend class INIFile;
    private:

        ///Name of the section. Taken from section header.
        miniini_private::c * Name;
        ///Length of the section, i.e. number of tags in it.
        miniini_private::ui Length;
        ///Tags in the section.
        miniini_private::c * * Tags;

        ///Data that doesn't need to be reinitialized in every ctor call.
        ///Temp tags buffer capacity.
        static miniini_private::ui temptagscap;
        ///Temp tags buffer - stores tags during loading.
        static miniini_private::c * * temptags;
        ///Tag buffer capacity.
        static miniini_private::ui tagcap;
        ///Temp tag buffer - stores currently processed tag during loading.
        static miniini_private::c * tagbuf;

    public:

        ///@return name of this section.
        const char * GetName() const;
        
        ///Read a string value from given tag.
        /**
         * Note that unlike other ReadXXX methods, ReadString only gives a 
         * pointer to data inside the INISection. If you want to use this
         * data after INIFile is destroyed, you have to copy it.
         * @param name Name of the tag to read from. Must be a valid zero terminated string.
         * @param out Reference to variable to read to.
         * @return true if the value was read succesfully
         * @return false if the tag does not exist (value was not read)
         */
        bool ReadString(const char * const name, const char * & out) const;

        ///Read an int value from given tag.
        /**
         * @param name Name of the tag to read from. Must be a valid zero terminated string.
         * @param out Reference to variable to read to.
         * @return true if the value was read succesfully
         * @return false if the tag does not exist (value was not read)
         */
        bool ReadInt(const char * const name, int & out) const;

        ///Read a float value from given tag.
        /**
         * @param name Name of the tag to read from. Must be a valid zero terminated string.
         * @param out Reference to variable to read to.
         * @return true if the value was read succesfully
         * @return false if the tag does not exist (value was not read)
         */
        bool ReadFloat(const char * const name, float & out) const;
        
        ///Read a bool value from given tag.
        /**
         * @param name Name of the tag to read from. Must be a valid zero terminated string.
         * @param out Reference to variable to read to.
         * @return true if the value was read succesfully
         * @return false if the tag does not exist (value was not read)
         */
        bool ReadBool(const char * const name, bool & out) const;

        ///Read an array of strings from a sequence of tags with given base name.
        /**
         * Reads values from tags named name1=, name2=, etc. where name is the
         * given base name for the tags. Continues to read successive elements
         * until it either reads as many elements as given capacity, there
         * are no more elements left in section, or it encounters an element
         * in wrong order, e.g. name4= after name2= .
         * Note that unlike other ReadXXXs methods, ReadStrings only gives a 
         * pointer to data inside the INISection. If you want to use this
         * data after INIFile is destroyed, you have to copy it.
         * @param name Base name of the tags to read from. Must be a valid zero terminated string.
         * @param out Pointer to buffer to read to.
         * @param cap Capacity of the buffer.
         * @return number of values read
         */
        unsigned ReadStrings(const char * const name, const char * * out, 
                             const unsigned cap) const;
        
        ///Read an array of ints from a sequence of tags with given base name.
        /**
         * Reads values from tags named name1=, name2=, etc. where name is the
         * given base name for the tags. Continues to read successive elements
         * until it either reads as many elements as given capacity, there
         * are no more elements left in section, or it encounters an element
         * in wrong order, e.g. name4= after name2= .
         * @param name Base name of the tags to read from. Must be a valid zero terminated string.
         * @param out Pointer to buffer to read to.
         * @param cap Capacity of the buffer.
         * @return number of values read
         */
        unsigned ReadInts(const char * const name, int * out, 
                          const unsigned cap) const;
        
        ///Read an array of floats from a sequence of tags with given base name.
        /**
         * Reads values from tags named name1=, name2=, etc. where name is the
         * given base name for the tags. Continues to read successive elements
         * until it either reads as many elements as given capacity, there
         * are no more elements left in section, or it encounters an element
         * in wrong order, e.g. name4= after name2= .
         * @param name Base name of the tags to read from. Must be a valid zero terminated string.
         * @param out Pointer to buffer to read to.
         * @param cap Capacity of the buffer.
         * @return number of values read
         */
        unsigned ReadFloats(const char * const name, float * out, 
                            const unsigned cap) const;

        ///Read an array of bools from a sequence of tags with given base name.
        /**
         * Reads values from tags named name1=, name2=, etc. where name is the
         * given base name for the tags. Continues to read successive elements
         * until it either reads as many elements as given capacity, there
         * are no more elements left in section, or it encounters an element
         * in wrong order, e.g. name4= after name2= .
         * @param name Base name of the tags to read from. Must be a valid zero terminated string.
         * @param out Pointer to buffer to read to.
         * @param cap Capacity of the buffer.
         * @return number of values read
         */
        unsigned ReadBools(const char * const name, bool * out, 
                           const unsigned cap) const;

        ///@return number of tags in the section.
        unsigned GetLength() const;
            
    private:

        ///Empty constructor.
        INISection();

        ///Destructor. 
        ~INISection();

        ///Loads the section from given position in a raw ini file buffer.
        /*
         * Loads the section and changes outlinetoken so that the caller can 
         * start with another section.
         * @param sectionname Name of the section to initialize
         * @param outlinetoken Start of the section in a raw ini file buffer.
         * @return true if the section is valid.
         * @return false if the section is empty.
         */
        bool Init(const miniini_private::c * const sectionname, 
                  miniini_private::c * * const outlinetoken);

        INISection(const INISection &);

        void operator = (const INISection &);

        ///Allocates/initializes static data
        static void InitTempData()
        {
            temptagscap = 8;
            temptags = new miniini_private::c * [temptagscap];
            tagcap = 64;
            tagbuf = new miniini_private::c [tagcap];
        }
        
        ///Destroys static data
        static void DestroyTempData()
        {
            temptagscap = 0;
            delete [] temptags;
            tagcap = 0;
            delete [] tagbuf;
        }
};

#endif
