// Copyright (C) 2009-2010 Ferdinand Majerech
// This file is part of MiniINI
// For conditions of distribution and use, see copyright notice in LICENSE.txt

#include <cstdio>
#include <cstring>

#include "typedefs.h"
#include "globals.h"
#include "log.h"
#include "linetoken.h"
#include "inisection.h"
#include "inifile.h"
    
using namespace miniini_private;

INIFile::INIFile()
    :Length(0)
    ,Sections(NULL)
{}

bool INIFile::OpenFile(const char * const fname)
{
    INISection::InitTempData();
    //opening file
    FILE * const file = fopen( fname, "rb" );
    if(file == NULL)
    {
        ERROR("Couldn't open file. Maybe it does not exist? File: %s",
              fname);
        return false;
    }
    //determining filesize
    i seek = fseek(file, 0, SEEK_END);
    if(seek != 0)
    {
        fclose(file);
        ERROR("Couldn't reach end of file. File: %s", fname);
        return false;
    }
    const i size = ftell(file);
    if(size == -1)
    {
        fclose(file);
        ERROR("Couldn't determine size of file. File: %s", fname);
        return false;
    }
    seek = fseek(file, 0, SEEK_SET);
    if(seek!=0)
    {
        fclose(file);
        ERROR("Couldn't reach start of file. File: %s", fname);
        return false;
    }
    //reading file to a buffer
    c * const buf = new c [size + 1];
    const i readnum = fread(buf, size, 1, file);
    //if 0 blocks succesfully read
    if(readnum == 0)
    {
        fclose(file);
        delete [] buf;
        ERROR("Could open but could not read from file. File might be "
              "corrupted or you might not have sufficient rights to "
              "read from it. File: %s", fname);
        return false;
    }
    //adding trailing zero to the buffer
    buf[size] = 0;
    //capacity of temporary buffer of pointers to sections
    ui tempsectionscap = 16;
    //temporary buffer of pointers to sections
    INISection * * tempsections = new INISection * [tempsectionscap]; 
    //pointer to start of current line if file buffer
    c * linetoken = strtok(buf, linesep );
    while( linetoken!=NULL )
    {
        //number of characters in this line (without trailing zero)
        const ui linesize = strlen(linetoken);
        //current part of the line we're in (name, header or val)
        LineToken linepart = LT_NAME;
        //currently processed character in this line is copied here
        c ch;
        //name of header (if any) in this line is stored here
        c * headername = new c[linesize - 1];
        ui hnamesize = 0;
        //going through characters in the line
        for(ui idx = 0; idx < linesize; ++idx)
        {
            ch = linetoken[idx];
            //if this character is a comment, skip rest of the line
            if(ch == comment)
            {
                if(linepart == LT_HEADER)
                {
                    WARNING("Part of a section header is commented out or there is a stray "
                            "'[' character. Ignoring. Line: %s File: %s", linetoken, fname);
                    linepart = LT_NAME;
                }
                break;
            }
            //if we're in name part of the line
            if(linepart == LT_NAME)
            {
                //whitespace is ignored
                if(ch == ' ' || ch == '\t')
                {
                    continue;
                }
                //if we hit '[' , we're probably in a section header
                if(ch == '[')
                {
                    linepart = LT_HEADER;
                    continue;
                }
            }
            //if this line is a header
            else if(linepart == LT_HEADER)
            {
                //whitespace is ignored
                if(ch == ' ' || ch == '\t')
                {
                    continue;
                }
                if(ch == ']')
                {
                    break;
                }
                //end of header not found; ignoring
                if(idx == linesize - 1)
                {
                    WARNING("'[' character present in a line but there's no ']'."
                            "This might be an incomplete header. Ignoring."
                            "Line: %s File: %s", linetoken, fname);
                    linepart = LT_NAME;    
                }
                //adding character to the header
                headername[hnamesize] = ch;
                ++hnamesize;
            }
        }
        //if this line is a header, we can start loading a section
        if(linepart == LT_HEADER)
        {
            if(!hnamesize)
            {
                WARNING("Header name empty. Ignoring. File: %s Line: %s",
                        fname, linetoken);
                delete [] headername; 
                linetoken = strtok( NULL, linesep );
                continue;
            }
            headername[hnamesize] = 0;
            //if temp section ptrs buffer runs out of space, reallocate it
            if(Length + 1 > tempsectionscap)
            {
                tempsectionscap *= 2;
                INISection * * const temptempsections = new INISection * [tempsectionscap];
                memcpy(temptempsections, tempsections, Length * sizeof(INISection *));
                delete [] tempsections;
                tempsections = temptempsections;
            }
            tempsections[Length] = new INISection();
            //if the section could not initialize correctly (e.g. is empty)
            if(!tempsections[Length]->Init(headername, &linetoken))
            {
                delete [] headername;
                delete tempsections[Length];
                continue;
            }
            delete [] headername;
            ++Length;
            //skipping strtok so we continue with header where section ctor finished
            continue;
        }
        delete [] headername;
        linetoken = strtok( NULL, linesep );
    }
    //delete file buffer now that everything is loaded
    delete [] buf;
    //copy pointers from tempsections to Sections and delete tempsections
    Sections = new INISection * [Length];
    memcpy(Sections, tempsections, Length * sizeof(INISection *));
    delete [] tempsections;
    fclose( file );
    INISection::DestroyTempData();
    return true;
}

INIFile::~INIFile()
{
    if(!Sections)
    {
        return;
    }
    for(ui section = 0; section < Length; ++section)
    {
        delete Sections[section];
    }
    delete [] Sections;
}

INISection * INIFile::GetSection(const char * const name) const
{
    INISection * const * sect = Sections;
    const INISection * const * const sectmax = Sections + Length;
    for(; sect < sectmax; ++sect)
    {
        if(!strcmp(name, (*sect)->Name))
        {
            return *sect;    
        }
    }
    ERROR("Missing requested section. Section: %s", name);
    return NULL;
}

INISection * INIFile::operator [] (const char * const name) const
{
    return GetSection(name);
}

unsigned INIFile::GetLength() const
{
    return static_cast<unsigned>(Length);
}

