// Copyright (C) 2009-2010 Ferdinand Majerech
// This file is part of MiniINI
// For conditions of distribution and use, see copyright notice in LICENSE.txt

#include <cstring>
#include <cerrno>
#include <cstdlib>

#include "typedefs.h"               
#include "globals.h"
#include "log.h"
#include "linetoken.h"
#include "inisection.h"

using namespace miniini_private;

ui INISection::temptagscap = 0;
c * * INISection::temptags = NULL;
ui INISection::tagcap = 0;
c * INISection::tagbuf = NULL;

INISection::INISection()
    :Name(NULL)
    ,Length(0)
    ,Tags(NULL)
{}

bool INISection::Init(const c * const sectionname, c * * const outlinetoken)
{
    //Copying name of the section
    const ui namelen = strlen(sectionname) + 1;
    Name = new c[namelen];
    memcpy(Name, sectionname, namelen);
    c * linetoken = *outlinetoken;
    //Current line token (points to start of current line in main file buffer)
    linetoken = strtok( NULL, linesep );
    //Size of temp tag buffer
    ui tagsize;
    //Currently processed character in the line is copied here
    c ch;
    //Ptr to currently processed character
    const c * chptr;
    //Ptr to end of current line
    const c * chmax;
    //Iterating through lines in the file
    while( linetoken != NULL )
    {
        tagsize = 0;
        //Number of characters in this line (without trailing zero character)
        const ui linesize = strlen(linetoken);
        //Current part of the line we're in (name, header or val)
        LineToken linepart = LT_NAME;
        //Temp name and val buffers ran out of space, reallocate them
        if(linesize >= tagcap)   
        {
            tagcap = linesize + 1;
            delete [] tagbuf;
            tagbuf = new c[tagcap];
        }
        //Iterating through characters in the line
        for(chptr = linetoken, chmax = linetoken + linesize; chptr < chmax; ++chptr)
        {
            ch = *chptr;
            //This character is a comment, skip rest of the line
            if(ch == comment)
            {
                //Comment following after [ but before ], if any
                if(linepart == LT_HEADER)
                {
                    linepart = LT_NAME;
                    WARNING("Part of a section header is commented out or there is a stray "
                            "'[' character. Ignoring. Section: %s Line: %s", Name, linetoken);
                }
                break;
            }
            //We're in name part of the line
            if(linepart == LT_NAME)
            {
                //Whitespace is ignored
                if(ch == ' ' || ch == '\t')
                {
                    continue;
                }
                //We are at '=' ; start of value part of the line
                if(ch == '=')
                {
                    //Name is empty, skip to next line
                    if(!tagsize)
                    {
                        WARNING("Tag with an empty name (no characters before '='. Ignoring."
                                "Section: %s Line: %s ", Name, linetoken);
                        goto NEXT_LINE;
                    }
                    //Name is not empty, start value part of the tag
                    tagbuf[tagsize] = 0;
                    ++tagsize;
                    linepart = LT_VAL;
                    continue;
                }
                //We hit '[', so this line is probably a header
                if(ch == '[')
                {
                    linepart = LT_HEADER;
                    continue;
                }
                //In any other case, append this character to name
                tagbuf[tagsize] = ch;
                ++tagsize;
            }
            //We're in value part of the line
            else if(linepart == LT_VAL)
            {
                //Whitespace is ignored
                if(ch == ' ' || ch == '\t')
                {
                    continue;
                }
                //If not whitespace, append this character to name
                tagbuf[tagsize] = ch;
                ++tagsize;
            }
            //This line is a header
            else if(linepart == LT_HEADER)
            {
                //Whitespace is ignored
                if(ch == ' ' || ch == '\t')
                {
                    continue;
                }
                //We're at ']', so we're at the end of header and skip rest of the line,
                //if any
                if(ch == ']')
                {
                    break;
                }
                //End of header not found; this is not a valid header ; ignoring
                if(chptr == chmax - 1)
                {
                    linepart = LT_NAME;
                    WARNING("'[' character present in a line but there's no ']'."
                            "This might be an incomplete header. Ignoring."
                            "Section: %s Line: %s ", Name, linetoken);
                    //We don't need to break as we're at the end of line
                }
            }
        }
        //This line is a header of another section, so we're finished reading this one
        if(linepart == LT_HEADER)
        {
            #ifdef INI_DEBUG
            if(tagsize)
            {
                WARNING("Unnecessary characters in a header line."
                        "Section: %s Line: %s ", Name, linetoken);
            }
            #endif
            break;//Skipping strtok
        }
        //This line is a tag with name and value
        else if(linepart == LT_VAL)
        {
            //Ignore this line if val is empty (last char is zero - that is, end of name)
            if(!tagbuf[tagsize - 1])
            {
                WARNING("Tag with an empty value (no characters after '='. Ignoring."
                        "Section: %s Line: %s ", Name, linetoken);
                goto NEXT_LINE;
            }
            //Add trailing zero
            tagbuf[tagsize] = 0;
            ++tagsize;
            //Reallocate temp lines buffer if not long enough
            if(Length >= temptagscap)
            {
                temptagscap *= 2;
                c * * const temptemptags = new c * [temptagscap];
                memcpy(temptemptags, temptags, Length * sizeof(c *));
                delete [] temptags;
                temptags = temptemptags;
            }
            //Add new tag to temp tag buffer
            temptags[Length] = new c [tagsize];
            memcpy(temptags[Length], tagbuf, tagsize * sizeof(c));
            ++Length;
        }
        //This line is garbage
        else if(tagsize > 1)
        {
            WARNING("Stray characters (nonempty line without tag, header or comments)"
                    "Section: %s Line: %s ", Name, linetoken);
        }
        NEXT_LINE:
        linetoken = strtok( NULL, linesep );
    }
    //Updating line token ptr of the caller
    *outlinetoken = linetoken;
    //There are no tags in the section
    if(!Length)
    {
        WARNING("Empty section. Ignoring. Section: %s", Name);
        delete [] Name;
        Name = NULL;
        return false;
    }
    //Copy temp lines buffer to Lines
    Tags = new c * [Length];
    memcpy(Tags, temptags, Length * sizeof(c *));
    return true;
}
    
INISection::~INISection()
{
    if(Name)
    {
        delete [] Name;
    }
    if(Tags)
    {
        for(ui line = 0; line < Length; ++line)
        {
            delete [] Tags[line];
        }
        delete [] Tags;
    }
}

const char * INISection::GetName() const
{
    return Name;
}

bool INISection::ReadString(const char * const name, const char * & out) const
{
    //Pointer to current tag
    c * const * tagptr = Tags;              
    //Pointer to end of tag buffer
    c * const * const maxtag = Tags + Length;
    //Pointer to current character in name
    register const u8 * nameptr; 
    //Pointer to current character in current tag name
    register u8 * tag;
    //Current character in name, current tag name
    register u8 namechar, tagchar;
    //Iterate through tags
    for(; tagptr < maxtag; ++tagptr)
    {
        //Comparing name with current tag name
        //Both are equal if their chars are same and they have same length
        nameptr = reinterpret_cast<const u8 *>(name); 
        tag = reinterpret_cast<u8 *>(*tagptr);
        //Iterate through chars in name and tag name as long as they are the same
        for(;;)
        {
            namechar = static_cast<u8>(*nameptr++);
            tagchar = static_cast<u8>(*(tag++));
            //We're at the end of name. If we're not also at the end of tag name,
            //this is not the tag we're looking for
            if (!namechar)
            { 
                if(tagchar) 
                {
                    goto NEXT_TAG;
                }       
                //Name and tag name have same chars and length:
                //We're at the end of tag name, so we've found requested tag
                else 
                {
                    break;
                }
            }
            //Chars in name and tag name are different: this is not the tag
            //we're looking for.
            if(namechar != tagchar)
            {
                goto NEXT_TAG;
            }
        }
        out = reinterpret_cast<char *>(tag);
        return true;
        NEXT_TAG:;
    }
    return false;
    
}

bool INISection::ReadInt(const char * const name, int & out) const
{
    //Points to char where strtol finished processing. Used to check for errors
    c * tail;
    errno = 0;
    const c * valstr;
    //Requested tag does not exist
    if(!ReadString(name, valstr))
    {
        return false;
    }
    const long tempout = strtol(valstr, &tail, 0);
    const c tailc = tail[0];
    if(tailc == valstr[0])
    {
        ERROR("Non-integer value in a tag where integer is expected."
              "Section: %s Tag: %s Value: %s", Name, name, valstr);
        return false;
    }
    if(errno)
    {
        //this is an error because in my tests strtol returned -1 on overflow.
        ERROR("Integer value out of range."
              "Section: %s Tag: %s Value: %s", Name, name, valstr);
        return false;
    }
    #ifdef INI_DEBUG
    if(tailc)
    {
        WARNING("Redunant characters in a tag where integer is "
                "expected. Reading integer."
                "Section: %s Tag: %s Value: %s", Name, name, valstr);
    }
    #endif
    out = static_cast<int>(tempout);
    return true;
}

bool INISection::ReadFloat(const char * const name, float & out) const
{
    //Points to char where strtof finished processing. Used to check for errors
    c * tail;
    errno = 0;
    const c * valstr;
    //Requested tag does not exist
    if(!ReadString(name, valstr))
    {
        return false;
    }
    const float tempout = strtof(valstr, &tail);
    const c tailc = tail[0];
    if(tailc == valstr[0])
    {
        ERROR("Non-float value in a tag where float is expected."
              "Section: %s Tag: %s Value: %s", Name, name, valstr);
        return false;
    }
    if(errno)
    {
        ERROR("Float value out of range."
              "Section: %s Tag: %s Value: %s", Name, name, valstr);
        return false;
    }
    #ifdef INI_DEBUG
    if(tailc)
    {
        WARNING("Redunant characters in a tag where float is "
                "expected. Reading float."
                "Section: %s Tag: %s Value: %s", Name, name, valstr);
    }
    #endif
    out = tempout;
    return true;
}

bool INISection::ReadBool( const char * const name, bool & out) const
{
    const c * valstr;
    //Requested tag does not exist
    if(!ReadString(name, valstr))
    {
        return false;
    }
    //Comparing string value with each true and false bool string 
    for(ui b = 0; b < BOOLNUM; ++b)
    {
        if(!strcmp(valstr, truebools[b]))
        {
            out = true;
            return true;
        }
        if(!strcmp(valstr, falsebools[b]))
        {
            out = false;
            return true;
        }
    }
    return false;
}

unsigned INISection::ReadStrings(const char * const name, const char * * out, 
                                 const unsigned cap) const
{
    //Pointer to current tag
    const c * const * tagptr = Tags;
    //Current tag
    const c * tag;
    //Pointer to the end of tags
    const c * const * const maxtagptr = Tags + Length;
    //Number of strings written to out
    ui elems = 0;
    const ui namelen = strlen(name);
    //Index of current element in ini (staring at 1, e.g. weapon1=, weapon2=)
    ui elemidx;
    //Points to char where strtol finished processing.
    //Strtol is used in parsing the index part of tag name.
    c * tail;
    for(; tagptr < maxtagptr; ++tagptr)
    {
        tag = *tagptr;
        //Tf tag name starts by name
        if(!strncmp(name, tag, namelen))
        {
            //Try to parse rest of tag name as number 
            elemidx = strtol(tag + namelen, &tail, 0);    
            const c tailc = tail[0];
            //Rest of the tag name is not an integer, probably different tag,
            //e.g. weapon= instead of weapon1=
            if(tailc == tag[namelen])
            {
                continue;
            }
            //Huge (out of range) integer in tag name- 
            //Probably a different tag name or error
            if(errno)
            {
                WARNING("Integer in tag name out of range. Ignoring. "
                        "Section: %s Tag name: %s", Name, tag);
                continue;
            }
            //Redunant chars after int - probably a different tag name or error 
            if(tailc)
            {
                WARNING("Redunant characters after integer in tag name. "
                        "Ignoring. Section: %s Tag name: %s", Name, tag);
                continue;
            }
            //Out of capacity  
            if(elemidx > cap || elemidx < 1)
            {
                WARNING("Integer in tag name out of range. Ignoring. Section: "
                        "%s Tag name: %s Range: 1-%d", Name, tag, cap);
                continue;
            }
            //Correct index at the end of name, reading array element.
            if(elemidx == elems + 1)
            {
                ++elems;
                out[elemidx - 1] = (tag) + strlen(tag) + 1;
            }
            //Incorrect order (e.g. a2 after a3) 
            else
            {
                WARNING("Array tag name in incorrect order (e.g. tag3= "
                        "after tag1= without tag2= in between) Ignoring "
                        "the tag. Section: %s Tag name: %s ", Name, tag);
            }
        }
    }
    return static_cast<unsigned>(elems);
}

unsigned INISection::ReadInts(const char * const name, int * out, 
                              const unsigned cap) const
{
    //Array of value strings to be converted to ints
    const c * * const valstrs = new const c * [cap];
    //Number of strings read by ReadStrings to valstrs
    const ui tempelems = ReadStrings(name, valstrs, cap);
    //Ptr to current string in valstrs
    const c * const * valstr = valstrs;
    //When valstr reaches this, we've iterated over all the strings
    const c * const * maxvalstr = valstrs + tempelems;
    //Number of actual valid ints written to out
    unsigned elems = 0;
    //Points to char where strtol finished processing. Used to check for errors
    c * tail;
    //Iterating through strings read by ReadStrings and converting them to ints
    for(; valstr < maxvalstr; ++valstr)
    {    
        errno = 0;
        const long tempelem = strtol(*valstr, &tail, 0);
        const c tailc = tail[0];
        if(tailc == *valstr[0])
        {
            ERROR("Non-integer value in an array tag where "
                  "integer is expected. Terminating array "
                  "reading. Section: %s Tag: %s%u Value: %s", 
                  Name, name, elems + 1, *valstr);
            break;
        }
        if(errno)
        {
            //this is an error because in some of my tests strtol returned -1 on overflow.
            ERROR("Integer value in an array tag out of range."
                  "Terminating array reading.Section: %s Tag: "
                  "%s%u Value: %s", Name, name, elems + 1, *valstr);
            break;
        }
        #ifdef INI_DEBUG
        if(tailc)
        {
            WARNING("Redunant characters in a tag where integer "
                    "is expected. Reading integer. Section: %s "
                    "Tag: %s%u Value: %s", Name, name, elems + 1,
                    *valstr);
        }
        #endif
        out[elems] = static_cast<int>(tempelem);   
        ++elems;
    }
    delete [] valstrs;
    return elems;
}       

unsigned INISection::ReadFloats(const char * const name, float * out, 
                                const unsigned cap) const
{
    //Array of value strings to be converted to floats
    const c * * const valstrs = new const c * [cap];
    //Number of strings read by ReadStrings to valstrs
    const ui tempelems = ReadStrings(name, valstrs, cap);
    //Ptr to current string in valstrs
    const c * const * valstr = valstrs;
    //When valstr reaches this, we've iterated over all the strings
    const c * const * const maxvalstr = valstrs + tempelems;
    //Number of actual valid floats written to out
    unsigned elems = 0;
    //Points to char where strtof finished processing. Used to check for errors
    c * tail;
    //Iterating through strings read by ReadStrings and converting them to floats
    for(; valstr < maxvalstr; ++valstr)
    {    
        errno = 0;
        const f tempelem = strtof(*valstr, &tail);
        const c tailc = tail[0];
        if(tailc == *valstr[0])
        {
            ERROR("Non-float value in an array tag where "
                  "float is expected. Terminating array "
                  "reading. Section: %s Tag: %s%u Value: %s", 
                  Name, name, elems + 1, *valstr);
            break;
        }
        if(errno)
        {
            ERROR("Float value in an array tag out of range."
                  "Terminating array reading.Section: %s Tag: "
                  "%s%u Value: %s", Name, name, elems + 1, *valstr);
            break;
        }
        #ifdef INI_DEBUG
        if(tailc)
        {
            WARNING("Redunant characters in a tag where float "
                    "is expected. Reading float. Section: %s "
                    "Tag: %s%u Value: %s", Name, name, elems + 1,
                    *valstr);
        }
        #endif
        out[elems] = tempelem;   
        ++elems;
    }
    delete [] valstrs;
    return elems;
}     

unsigned INISection::ReadBools(const char * const name, bool * out, 
                               const unsigned cap) const
{
    //Array of value strings to be converted to bools 
    const c * * const valstrs = new const c * [cap];
    //Number of strings read by ReadStrings to valstrs
    const ui tempelems = ReadStrings(name, valstrs, cap);
    //Ptr to current string in valstrs
    const c * const * valstr = valstrs;
    //When valstr reaches this, we've iterated over all the strings
    const c * const * const maxvalstr = valstrs + tempelems;
    //Number of actual valid bools written to out
    unsigned elems = 0;
    //Iterating through strings read by ReadStrings and converting them to bools
    for(; valstr < maxvalstr; ++valstr, ++elems)
    {
        //Comparing string value with each true and false bool string 
        for(ui b = 0; b < BOOLNUM; ++b)
        {
            if(!strcmp(*valstr, truebools[b]))
            {
                out[elems] = true;
                goto NEXT_ELEM;
            }
            if(!strcmp(*valstr, falsebools[b]))
            {
                out[elems] = false;
                goto NEXT_ELEM;
            }
        }
        //If we reach this, string didn't match any bool string
        WARNING("Non-bool value in an array tag where bool is expected. "
                "Terminating array reading. Section: %s Tag: %s%u "
                "Value: %s", Name, name, elems + 1, *valstr );
        break;
        NEXT_ELEM:;
    }
    delete [] valstrs;
    return elems;
}

unsigned INISection::GetLength() const
{
    return static_cast<unsigned>(Length);
}

