// Copyright (C) 2009-2010 Ferdinand Majerech
// This file is part of MiniINI
// For conditions of distribution and use, see copyright notice in LICENSE.txt

#include <cstdio>
#include <cstring>

#include "../../miniini.h"

bool ReadString(INISection * section)
{
    const char * readstring;
    if(!section->ReadString("valstring", readstring))
    {
        puts("Missing tag valstring= or error reading value");
        return false;
    }
    if(strcmp(readstring, "string"))
    {
        printf("Didn't read [%s] valstring= as \"string\" \n valstring was "
               "read as %s\n", section->GetName(), readstring);
        return false;
    }
    return true;
}

bool ReadInt(INISection * section)
{
    int readint;
    if(!section->ReadInt("valint", readint))
    {
        puts("Missing tag valint= or error reading value");
        return false;
    }
    if(readint != -5)
    {
        printf("Didn't read [%s] valint= as -5 \n valint was read as %i\n"
                , section->GetName(), readint);
        return false;
    }
    return true;
}

bool ReadFloat(INISection * section)
{
    float readfloat;
    if(!section->ReadFloat("valfloat", readfloat))
    {
        puts("Missing tag valfloat= or error reading value");
        return false;
    }
    if(readfloat != 1.5f)
    {
        printf("Didn't read [%s] valfloat= as 1.5 \n valfloat was read as %f\n"
                , section->GetName(), readfloat);
        return false;
    }
    return true;
}

bool ReadBool(INISection * section)
{
    bool readbool;
    if(!section->ReadBool("valbool", readbool))
    {
        puts("Missing tag valbool= or error reading value");
        return false;
    }
    if(readbool != true)
    {
        printf("Didn't read [%s] valbool= as true \n valbool was read as %s\n"
                , section->GetName(), readbool ? "true" : "false");
        return false;
    }
    return true;
}

bool ReadStrings(INISection * section)
{
    int stringcount;
    if(!section->ReadInt("stringcount", stringcount))
    {
        puts("Missing tag stringcount= or error reading value");
        return false;
    }
    const char * * readstrings = new const char * [stringcount];
    int readcount = section->ReadStrings("string", readstrings, stringcount);
    if( readcount != stringcount)
    {
	    printf("ReadStrings didn't read stringcount= elements\n");
	    printf("stringcount= : %d elements read : %u\n", stringcount, stringcount);
        return false;
    }
    delete [] readstrings;
    return true;
}

bool ReadInts(INISection * section)
{
    int intcount;
    if(!section->ReadInt("intcount", intcount))
    {
        puts("Missing tag intcount= or error reading value");
        return false;
    }
    int * readints = new int [intcount];
    int readcount = section->ReadInts("int", readints, intcount);
    if( readcount != intcount)
    {
	    printf("ReadInts didn't read intcount= elements\n");
	    printf("intcount= : %d elements read : %u\n", intcount, readcount);
        return false;
    }
    delete [] readints;
    return true;
}

bool ReadFloats(INISection * section)
{
    int floatcount;
    if(!section->ReadInt("floatcount", floatcount))
    {
        puts("Missing tag intcount= or error reading value");
        return false;
    }
    float * readfloats = new float [floatcount];
    int readcount = section->ReadFloats("float", readfloats, floatcount);
    if( readcount != floatcount)
    {
	    printf("ReadFloats didn't read floatcount= elements\n");
	    printf("floatcount= : %d elements read : %u\n", floatcount, readcount);
        return false;
    }
    delete [] readfloats;
    return true;
}

bool ReadBools(INISection * section)
{
    int boolcount;
    if(!section->ReadInt("boolcount", boolcount))
    {
        puts("Missing tag intcount= or error reading value");
        return false;
    }
    bool * readbools = new bool [boolcount];
    int readcount = section->ReadBools("bool", readbools, boolcount);
    if( readcount != boolcount)
    {
	    printf("ReadBools didn't read boolcount= elements\n");
	    printf("boolcount= : %d elements read : %u\n", boolcount, readcount);
        return false;
    }
    delete [] readbools;
    return true;
}

bool TagCount(INISection * section)
{
    int numtags;
    if(!section->ReadInt("numtags", numtags))
    {
        puts("Missing tag numtags= or error reading value");
        return false;
    }
    int tagsloaded = section->GetLength();
    if(numtags != tagsloaded)
    {
        printf("numtags= in doesn't match with number of tags actually "
               "loaded. numtags= : %d tags loaded : %d\n", numtags, tagsloaded);
        return false;
    }
    return true;
}

bool Spaces(INISection * section)
{
    const char * spacesinaval;
    if(!section->ReadString("spacesinaname", spacesinaval))
    {
        puts("Missing tag spacesinaname= or error reading value");
        return false;
    }
    if(strcmp(spacesinaval, "spacesinaval"))
    {
        puts("Whitespaces do not seem to be ignored correctly");
        return false;
    }
    return true;
}

class TestSuite
{
    private:
        char * Section;
        bool(**Tests )(INISection *);
        char (* TestNames) [64];
        INIFile & INI;
        
    public:
        TestSuite(char * section, char (* testnames) [64],
                  bool(**tests )(INISection *), INIFile & ini)
            :Section(section)
            ,Tests(tests)
            ,TestNames(testnames)
            ,INI(ini)
        {}

        void Run()
        {
            printf("Starting run of test suite [%s]...\n", Section);
            INISection * section = INI.GetSection(Section);
            if(!section)
            {
                printf("ERROR: Missing section %s: can't run test suite\n",
                        Section);
                return;
            }
            unsigned successes = 0;
            unsigned failures = 0;
            bool(**test)(INISection *) = Tests;
            char (* name) [64] = &TestNames[0];
            for(;*test; ++test, ++name)
            {
                puts("Running test:");
                puts(*name);
                if((*test)(section))
                {
                    ++successes;
                    puts("Test succeeded");
                }
                else
                {
                    ++failures;
                    puts("Test FAILED");
                }
            }
            printf("Test suite %s: %u tests: Successes: %u Failures: %u\n\n", 
                   Section, successes + failures, successes, failures);
        }

};

void RunTests(INIFile & ini)
{         
    puts("Testing count of loaded sections...");
    INISection * section = ini.GetSection("SETTINGS");
    if(!section)
    {
        printf("ERROR: Missing section SETTINGS: can't run ANY test suite\n");
        return;
    }
    int sections;
    if(!section->ReadInt("sections", sections))
    {
        puts("ERROR: Missing tag sections= in [SETTINGS] section or error "
             "reading value: can't run ANY test suite");
        return;
    }
    int sectionsloaded = ini.GetLength();
    if(sections != sectionsloaded)
    {
        printf("ERROR: sections= doesn't match with number of sections "
               "loaded. sections=: %d Sections loaded: %d \ncan't run ANY "
               "test suite", sections, sectionsloaded);
        return;
    }
    puts("Count of loaded sections test succeeded\n");
    
    char valtypes_name [] = "valtypes";
    bool(*valtypes_tests [])(INISection *) = {&ReadInt, &ReadFloat, &ReadBool, 
                                              &ReadString, NULL};
    char valtypes_testnames [] [64] = {"ReadInt", "ReadFloat", "ReadBool", 
                                       "ReadString"};

    char arrays_name [] = "arrays";
    bool(*arrays_tests [])(INISection *) = {&ReadInts, &ReadFloats, &ReadBools,
                                            &ReadStrings, NULL};
    char arrays_testnames [] [64] = {"ReadInts", "ReadFloats", "ReadBools",
                                     "ReadStrings"};
    char garbage_name [] = "garbage";
    bool(*garbage_tests [])(INISection *) = {&TagCount, &Spaces, NULL};
    char garbage_testnames [] [64] = {"TagCount", "Spaces"};
        
    TestSuite ValTypes(valtypes_name, valtypes_testnames, valtypes_tests, ini);
    TestSuite Arrays(arrays_name, arrays_testnames, arrays_tests, ini);
    TestSuite Garbage(garbage_name, garbage_testnames, garbage_tests, ini);
    ValTypes.Run();
    Arrays.Run();
    Garbage.Run();
}

int main()
{
    char fname [] = "test/test.ini";
    puts("INI regression test: Loading test file:");
    puts(fname);
    INIFile ini;
    if(!ini.OpenFile(fname))
    {
        puts("ERROR:\nCould not open file");
    }
    RunTests(ini);
    return 0;
}         
